@@ -19,7 +19,7 @@
  */
 
 #ifndef VERSION
-# define VERSION "1.1"
+# define VERSION "1.2"
 #endif
 
 #ifndef EXTRA_VERSION
@@ -45,6 +45,8 @@
 #include <sys/time.h>
 #include <sys/stat.h>
 
+#define HAVE_GETOPT_LONG_ONLY
+
 #ifdef __linux__
 # include <sys/ioctl.h>
 # include <sys/mount.h>
@@ -56,6 +58,7 @@
 # define HAVE_DIRECT_IO
 # define HAVE_LINUX_ASYNC_IO
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 # define MAX_RW_COUNT		0x7ffff000 /* 2G - 4K */
 #endif
 
@@ -65,6 +68,7 @@
 # define HAVE_POSIX_MEMALIGN
 # define HAVE_MKOSTEMP
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 #endif
 
 #if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
@@ -75,6 +79,7 @@
 # define HAVE_MKOSTEMP
 # define HAVE_DIRECT_IO
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 #endif
 
 #ifdef __DragonFly__
@@ -82,6 +87,7 @@
 # define HAVE_CLOCK_GETTIME
 # define HAVE_MKOSTEMP
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 #endif
 
 #ifdef __OpenBSD__
@@ -94,6 +100,7 @@
 # define HAVE_POSIX_MEMALIGN
 # define HAVE_MKOSTEMP
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 #endif
 
 #ifdef __APPLE__ /* OS X */
@@ -103,15 +110,16 @@
 # include <sys/uio.h>
 # define HAVE_NOCACHE_IO
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 #endif
 
 #ifdef __sun	/* Solaris */
 # include <sys/dkio.h>
 # include <sys/vtoc.h>
 # define HAVE_CLOCK_GETTIME
-# define HAVE_DIRECT_IO
-# define O_DIRECT	O_DSYNC
+# define HAVE_POSIX_FADVICE
 # define HAVE_ERR_INCLUDE
+# define HAVE_STATVFS
 #endif
 
 #ifdef __MINGW32__ /* Windows */
@@ -126,6 +134,10 @@
 # define HAVE_POSIX_FDATASYNC
 #endif
 
+#ifdef HAVE_STATVFS
+# include <sys/statvfs.h>
+#endif
+
 #ifdef HAVE_ERR_INCLUDE
 # include <err.h>
 #else
@@ -167,6 +179,7 @@ void warnx(const char *fmt, ...)
 #endif /* HAVE_ERR_INCLUDE */
 
 #define NSEC_PER_SEC	1000000000ll
+#define USEC_PER_SEC	1000000L
 
 #ifdef HAVE_CLOCK_GETTIME
 
@@ -180,6 +193,16 @@ static inline long long now(void)
 	return ts.tv_sec * NSEC_PER_SEC + ts.tv_nsec;
 }
 
+static inline double timestamp(void)
+{
+	struct timespec ts;
+
+	if (clock_gettime(CLOCK_REALTIME, &ts))
+		err(3, "clock_gettime failed");
+
+	return ts.tv_sec + (double)ts.tv_nsec / NSEC_PER_SEC;
+}
+
 #else
 
 static inline long long now(void)
@@ -192,6 +215,16 @@ static inline long long now(void)
 	return tv.tv_sec * NSEC_PER_SEC + tv.tv_usec * 1000ll;
 }
 
+static inline double timestamp(void)
+{
+	struct timeval tv;
+
+	if (gettimeofday(&tv, NULL))
+		err(3, "gettimeofday failed");
+
+	return tv.tv_sec + (double)tv.tv_usec / USEC_PER_SEC;
+}
+
 #endif /* HAVE_CLOCK_GETTIME */
 
 #ifndef HAVE_MKOSTEMP
@@ -297,42 +330,6 @@ int fdatasync(int fd)
 }
 #endif
 
-void usage(void)
-{
-	fprintf(stderr,
-			" Usage: ioping [-ABCDRLWYykq] [-c count] [-i interval] [-s size] [-S wsize]\n"
-			"               [-o offset] [-w deadline] [-pP period] directory|file|device\n"
-			"        ioping -h | -v\n"
-			"\n"
-			"      -c <count>      stop after <count> requests\n"
-			"      -i <interval>   interval between requests (1s)\n"
-			"      -l <speed>      speed limit in bytes per second\n"
-			"      -t <time>       minimal valid request time (0us)\n"
-			"      -T <time>       maximum valid request time\n"
-			"      -s <size>       request size (4k)\n"
-			"      -S <wsize>      working set size (1m)\n"
-			"      -o <offset>     working set offset (0)\n"
-			"      -w <deadline>   stop after <deadline> time passed\n"
-			"      -p <period>     print raw statistics for every <period> requests\n"
-			"      -P <period>     print raw statistics for every <period> in time\n"
-			"      -A              use asynchronous I/O\n"
-			"      -C              use cached I/O (no cache flush/drop)\n"
-			"      -B              print final statistics in raw format\n"
-			"      -D              use direct I/O (O_DIRECT)\n"
-			"      -R              seek rate test\n"
-			"      -L              use sequential operations\n"
-			"      -W              use write I/O (please read manpage)\n"
-			"      -G              read-write ping-pong mode\n"
-			"      -Y              use sync I/O (O_SYNC)\n"
-			"      -y              use data sync I/O (O_DSYNC)\n"
-			"      -k              keep and reuse temporary file (ioping.tmp)\n"
-			"      -q              suppress human-readable output\n"
-			"      -h              display this message and exit\n"
-			"      -v              display version and exit\n"
-			"\n"
-	       );
-}
-
 void version(void)
 {
 	fprintf(stderr, "ioping %s\n", VERSION EXTRA_VERSION);
@@ -397,8 +394,8 @@ static struct suffix time_suffix[] = {
 	{ NULL,		0ll },
 };
 
-long long parse_suffix(const char *str, struct suffix *sfx,
-		       long long min, long long max)
+double parse_suffix(const char *str, struct suffix *sfx,
+		    long long min, long long max)
 {
 	char *end;
 	double val, den;
@@ -481,7 +478,7 @@ void print_time(long long val)
 char *path = NULL;
 char *fstype = "";
 char *device = "";
-off_t device_size = 0;
+long long device_size = 0;
 
 int fd;
 void *buf;
@@ -504,6 +501,7 @@ long long interval = NSEC_PER_SEC;
 struct timespec interval_ts;
 long long deadline = 0;
 long long speed_limit = 0;
+double rate_limit = 0;
 
 long long min_valid_time = 0;
 long long max_valid_time = LLONG_MAX;
@@ -518,10 +516,106 @@ int keep_file = 0;
 off_t offset = 0;
 off_t woffset = 0;
 
+long long request = 0;
+long long warmup_request = 1;
 long long stop_at_request = 0;
 
+int json = 0;
+int json_line = 0;
+
 int exiting = 0;
 
+const char *options = "hvkALRDCWGYBqyi:t:T:w:s:S:c:o:p:P:l:r:a:J";
+
+#ifdef HAVE_GETOPT_LONG_ONLY
+
+static struct option long_options[] = {
+	{"help",	no_argument,		NULL,	'h'},
+	{"version",	no_argument,		NULL,	'v'},
+
+	{"keep",	no_argument,		NULL,	'k'},
+
+	{"quiet",	no_argument,		NULL,	'q'},
+	{"batch",	no_argument,		NULL,	'B'},
+	{"json",	no_argument,		NULL,	'J'},
+
+	{"rapid",	no_argument,		NULL,	'R'},
+	{"linear",	no_argument,		NULL,	'L'},
+	{"direct",	no_argument,		NULL,	'D'},
+	{"cached",	no_argument,		NULL,	'C'},
+	{"sync",	no_argument,		NULL,	'Y'},
+	{"dsync",	no_argument,		NULL,	'y'},
+	{"async",	no_argument,		NULL,	'A'},
+	{"write",	no_argument,		NULL,	'W'},
+	{"read-write",	no_argument,		NULL,	'G'},
+
+	{"size",	required_argument,	NULL,	's'},
+	{"work-size",	required_argument,	NULL,	'S'},
+	{"work-offset",	required_argument,	NULL,	'o'},
+
+	{"count",	required_argument,	NULL,	'c'},
+	{"work-time",	required_argument,	NULL,	'w'},
+
+	{"interval",	required_argument,	NULL,	'i'},
+	{"speed-limit",	required_argument,	NULL,	'l'},
+	{"rate-limit",  required_argument,	NULL,	'r'},
+
+	{"warmup",	required_argument,	NULL,	'a'},
+	{"min-time",	required_argument,	NULL,	't'},
+	{"max-time",	required_argument,	NULL,	'T'},
+
+	{"print-count", required_argument,	NULL,	'p'},
+	{"print-interval", required_argument,	NULL,	'P'},
+
+	{0,		0,			NULL,	0},
+};
+
+#endif /* HAVE_GETOPT_LONG */
+
+void usage(void)
+{
+	fprintf(stderr,
+			" Usage: ioping [-ABCDRLWYykq] [-c count] [-i interval] [-s size] [-S wsize]\n"
+			"               [-o offset] [-w deadline] [-pP period] directory|file|device\n"
+			"        ioping -h | -v\n"
+			"\n"
+			" options:\n"
+			"      -A, -async                 use asynchronous I/O\n"
+			"      -C, -cached                use cached I/O (no cache flush/drop)\n"
+			"      -D, -direct                use direct I/O (O_DIRECT)\n"
+			"      -G, -read-write            read-write ping-pong mode\n"
+			"      -L, -linear                use sequential operations\n"
+			"      -W, -write                 use write I/O (please read manpage)\n"
+			"      -Y, -sync                  use sync I/O (O_SYNC)\n"
+			"      -y, -dsync                 use data sync I/O (O_DSYNC)\n"
+			"      -R, -rapid                 test with rapid I/O during 3s (-q -i 0 -w 3)\n"
+			"      -k, -keep                  keep and reuse temporary file (ioping.tmp)\n"
+			"\n"
+			" parameters:\n"
+			"      -a, -warmup <count>        ignore <count> first requests (1)\n"
+			"      -c, -count <count>         stop after <count> requests\n"
+			"      -i, -interval <time>       interval between requests (1s)\n"
+			"      -s, -size <size>           request size (4k)\n"
+			"      -S, -work-size <size>      working set size (1m)\n"
+			"      -o, -work-offset <size>    working set offset (0)\n"
+			"      -w, -work-time <time>      stop after <time> passed\n"
+			"      -l, -speed-limit <size>    limit speed with <size> per second\n"
+			"      -r, -rate-limit <count>    limit rate with <count> per second\n"
+			"\n"
+			" output:\n"
+			"      -p, -print-count <count>   print raw statistics for every <count> requests\n"
+			"      -P, -print-interval <time> print raw statistics for every <time>\n"
+			"      -t, -min-time <time>       minimal valid request time (0us)\n"
+			"      -T, -max-time <time>       maximum valid request time\n"
+			"      -B, -batch                 print final statistics in raw format\n"
+			"      -J, -json                  print output in JSON format\n"
+			"      -q, -quiet                 suppress human-readable output\n"
+			"      -h, -help                  display this message and exit\n"
+			"      -v, -version               display version and exit\n"
+			"\n"
+	       );
+}
+
 void parse_options(int argc, char **argv)
 {
 	int opt;
@@ -531,7 +625,14 @@ void parse_options(int argc, char **argv)
 		exit(1);
 	}
 
-	while ((opt = getopt(argc, argv, "hvkALRDCWGYBqyi:t:T:w:s:S:c:o:p:P:l:")) != -1) {
+	while ((opt =
+#ifdef HAVE_GETOPT_LONG_ONLY
+		getopt_long_only(argc, argv, options, long_options, NULL)
+#else
+		getopt(argc, argv, options)
+#endif
+	) != -1) {
+
 		switch (opt) {
 			case 'h':
 				usage();
@@ -544,8 +645,15 @@ void parse_options(int argc, char **argv)
 				default_size = 1<<18;
 				break;
 			case 'l':
+				if (!custom_interval)
+					interval = 0;
 				speed_limit = parse_size(optarg);
 				break;
+			case 'r':
+				if (!custom_interval)
+					interval = 0;
+				rate_limit = parse_suffix(optarg, int_suffix, 0, NSEC_PER_SEC);
+				break;
 			case 'R':
 				if (!custom_interval)
 					interval = 0;
@@ -612,9 +720,15 @@ void parse_options(int argc, char **argv)
 				quiet = 1;
 				batch_mode = 1;
 				break;
+			case 'J':
+				json = 1;
+				break;
 			case 'c':
 				stop_at_request = parse_int(optarg);
 				break;
+			case 'a':
+				warmup_request = parse_int(optarg);
+				break;
 			case 'k':
 				keep_file = 1;
 				break;
@@ -1086,9 +1200,11 @@ static void start_statistics(struct statistics *s, unsigned long long start) {
 	s->start = start;
 }
 
-static void add_statistics(struct statistics *s, long long val) {
+static int add_statistics(struct statistics *s, long long val) {
 	s->count++;
-	if (val < min_valid_time) {
+	if (request <= warmup_request) {
+		/* warmup */
+	} else if (val < min_valid_time) {
 		s->too_fast++;
 	} else if (val > max_valid_time) {
 		s->too_slow++;
@@ -1100,7 +1216,10 @@ static void add_statistics(struct statistics *s, long long val) {
 			s->min = val;
 		if (val > s->max)
 			s->max = val;
+		return 1;
 	}
+
+	return 0;
 }
 
 static void merge_statistics(struct statistics *s, struct statistics *o) {
@@ -1152,32 +1271,126 @@ static void dump_statistics(struct statistics *s) {
 			(unsigned long)s->load_time);
 }
 
+static void json_request(size_t io_size, long long io_time, int valid)
+{
+	printf("%s{\n"
+	       "  \"timestamp\": %f,\n"
+	       "  \"target\": {\n"
+	       "    \"path\": \"%s\",\n"
+	       "    \"fstype\": \"%s\",\n"
+	       "    \"device\": \"%s\",\n"
+	       "    \"device_size\": %ld\n"
+	       "  },\n"
+	       "  \"io\": {\n"
+	       "    \"request\": %ld,\n"
+	       "    \"operation\": \"%s\",\n"
+	       "    \"size\": %ld,\n"
+	       "    \"time\": %lu,\n"
+	       "    \"ignored\": %s\n"
+	       "  }\n"
+	       "}",
+	       json_line++ ? "," : "",
+	       timestamp(),
+	       path,
+	       fstype,
+	       device,
+	       (long)device_size,
+	       (long)request,
+	       write_test ? "write" : "read",
+	       (unsigned long)io_size,
+	       (unsigned long)io_time,
+	       valid ? "false" : "true");
+	fflush(stdout);
+}
+
+static void json_statistics(struct statistics *s)
+{
+	printf("%s{\n"
+	       "  \"timestamp\": %f,\n"
+	       "  \"target\": {\n"
+	       "    \"path\": \"%s\",\n"
+	       "    \"fstype\": \"%s\",\n"
+	       "    \"device\": \"%s\",\n"
+	       "    \"device_size\": %ld\n"
+	       "  },\n"
+	       "  \"stat\": {\n"
+	       "    \"count\": %lu,\n"
+	       "    \"size\": %lu,\n"
+	       "    \"time\": %.0f,\n"
+	       "    \"iops\": %f,\n"
+	       "    \"bps\": %.0f,\n"
+	       "    \"min\": %lu,\n"
+	       "    \"avg\": %.0f,\n"
+	       "    \"max\": %lu,\n"
+	       "    \"mdev\": %.0f\n"
+	       "  },\n"
+	       "  \"load\": {\n"
+	       "    \"count\": %lu,\n"
+	       "    \"size\": %lu,\n"
+	       "    \"time\": %lu,\n"
+	       "    \"iops\": %f,\n"
+	       "    \"bps\": %.0f\n"
+	       "  }\n"
+	       "}",
+	       json_line++ ? "," : "",
+	       timestamp(),
+	       path,
+	       fstype,
+	       device,
+	       (long)device_size,
+	       (unsigned long)s->valid,
+	       (unsigned long)s->size,
+	       s->sum,
+	       s->iops,
+	       s->speed,
+	       (unsigned long)s->min,
+	       s->avg,
+	       (unsigned long)s->max,
+	       s->mdev,
+	       (unsigned long)s->count,
+	       (unsigned long)s->load_size,
+	       (unsigned long)s->load_time,
+	       s->load_iops,
+	       s->load_speed);
+	fflush(stdout);
+}
+
 int main (int argc, char **argv)
 {
 	ssize_t ret_size;
 	struct stat st;
+	int valid;
 	int ret;
 
 	struct statistics part, total;
 
-	long long request, this_time;
+	long long this_time;
 	long long time_now, time_next, period_deadline;
 
-	setvbuf(stdout, NULL, _IOLBF, 0);
-
 	parse_options(argc, argv);
 
+	if (!json)
+		setvbuf(stdout, NULL, _IOLBF, 0);
+
 	if (!size)
 		size = default_size;
 
 	if (size <= 0)
 		errx(1, "request size must be greater than zero");
 
-	if (custom_interval && speed_limit)
-		errx(1, "speed limit and interval cannot be set simultaneously");
+	if (speed_limit) {
+		long long i = size * NSEC_PER_SEC / speed_limit;
 
-	if (speed_limit)
-		interval = size * NSEC_PER_SEC / speed_limit;
+		if (i > interval)
+			interval = i;
+	}
+
+	if (rate_limit) {
+		long long i = NSEC_PER_SEC / rate_limit;
+
+		if (i > interval)
+			interval = i;
+	}
 
 #ifdef MAX_RW_COUNT
 	if (size > MAX_RW_COUNT)
@@ -1249,7 +1462,7 @@ int main (int argc, char **argv)
 		} else {
 			device_size = st.st_size;
 			fstype = "block";
-			device = "device ";
+			device = "device";
 		}
 	} else {
 		errx(2, "unsupported destination: \"%s\"", path);
@@ -1304,6 +1517,15 @@ skip_preparation:
 			err(2, "failed to open \"%s\"", path);
 	}
 
+#ifdef HAVE_STATVFS
+	if (S_ISDIR(st.st_mode) || S_ISREG(st.st_mode)) {
+		struct statvfs vfs;
+
+		if (!fstatvfs(fd, &vfs))
+			device_size = (long long)vfs.f_frsize * vfs.f_blocks;
+	}
+#endif
+
 	/* No readahead for non-cached I/O, we'll invalidate it anyway */
 	if (randomize || !cached) {
 #ifdef HAVE_POSIX_FADVICE
@@ -1325,7 +1547,6 @@ skip_preparation:
 
 	set_signal();
 
-	request = 0;
 	woffset = 0;
 
 	time_now = now();
@@ -1333,6 +1554,9 @@ skip_preparation:
 	start_statistics(&part, time_now);
 	start_statistics(&total, time_now);
 
+	if (json)
+		printf("[");
+
 	if (deadline)
 		deadline += time_now;
 
@@ -1384,22 +1608,20 @@ skip_preparation:
 
 		this_time = time_now - this_time;
 
-		if (request == 1) {
-			/* warmup */
-			part.count++;
-		} else {
-			add_statistics(&part, this_time);
-		}
+		valid = add_statistics(&part, this_time);
 
-		if (!quiet) {
+		if (quiet) {
+			/* silence */
+		} else if (json) {
+			json_request(ret_size, this_time, valid);
+		} else {
 			print_size(ret_size);
-			printf(" %s %s (%s %s", write_test ? ">>>" : "<<<",
+			printf(" %s %s (%s %s ", write_test ? ">>>" : "<<<",
 					path, fstype, device);
-			if (device_size)
-				print_size(device_size);
+			print_size(device_size);
 			printf("): request=%lu time=", (long unsigned)request);
 			print_time(this_time);
-			if (request == 1) {
+			if (request <= warmup_request) {
 				printf(" (warmup)");
 			} else if (this_time < min_valid_time) {
 				printf(" (too fast)");
@@ -1416,10 +1638,13 @@ skip_preparation:
 			printf("\n");
 		}
 
-		if ((period_request && (part.count >= period_request)) ||
+		if ((period_request && (part.valid >= period_request)) ||
 		    (period_time && (time_next >= period_deadline))) {
 			finish_statistics(&part, time_now);
-			dump_statistics(&part);
+			if (json)
+				json_statistics(&part);
+			else
+				dump_statistics(&part);
 			merge_statistics(&total, &part);
 			start_statistics(&part, time_now);
 			period_deadline = time_now + period_time;
@@ -1454,6 +1679,12 @@ skip_preparation:
 	merge_statistics(&total, &part);
 	finish_statistics(&total, time_now);
 
+	if (json) {
+		json_statistics(&total);
+		printf("]\n");
+		return 0;
+	}
+
 	if (batch_mode) {
 		dump_statistics(&total);
 		return 0;
@@ -1462,9 +1693,8 @@ skip_preparation:
 	if (quiet && (period_time || period_request))
 		return 0;
 
-	printf("\n--- %s (%s %s", path, fstype, device);
-	if (device_size)
-		print_size(device_size);
+	printf("\n--- %s (%s %s ", path, fstype, device);
+	print_size(device_size);
 	printf(") ioping statistics ---\n");
 	print_int(total.valid);
 	printf(" requests completed in ");
